import attr
from attr.validators import instance_of, deep_iterable

@attr.s(kw_only=True)
class Network:
    name = attr.ib(validator=instance_of(str))
    id = attr.ib(validator=instance_of(str))

@attr.s(kw_only=True)
class Options:
    network = attr.ib(validator=instance_of(Network))
    timeout = attr.ib(validator=instance_of((float, int)))
    remove = attr.ib(validator=instance_of(bool))
    run_dir = attr.ib(validator=instance_of(str))
    build = attr.ib(validator=deep_iterable(member_validator=instance_of(str)))

class AgentStatus:
    NULL = 'null'
    IN_PROGRESS = 'in-progress'
    STARTED = 'started'
    FAILED = 'failed'
    STOPPED = 'stopped'

class RunCondition:
    # Actions
    CREATE = 'create'
    START = 'start'
    PRE_START = 'pre-start'
    POST_START = 'post-start'
    PING = 'ping'
    # States
    NULL = 'null'
    BUILD_IMAGE = 'build-image'
    STARTED = 'started'
    RUNNING = 'running'
    FAILED = 'failed'

    def __init__(self):
        self.actions = []
        self.state = self.NULL

    def already_running(self):
        self.state = self.RUNNING

    def pinged(self):
        self.actions.append(self.PING)
        self.state = self.RUNNING

    def pre_started(self):
        self.actions.append(self.PRE_START)

    def post_started(self):
        self.actions.append(self.POST_START)

    def build_image(self):
        self.actions.append(self.BUILD_IMAGE)

    def started(self):
        self.state = self.STARTED
        self.actions.append(self.START)

    def fail(self):
        self.state = self.FAILED

class Actions:
    START = 'start'
    STOP = 'stop'

group_name = None

def set_group_name(name):
    global group_name
    group_name = name

def _unset_group_name():
    global group_name
    group_name = None



import os
import click

from miniboss import services
from miniboss.exceptions import MinibossCLIError

@click.group()
def cli():
    pass

def get_main_directory():
    """Return the path to the directory where the main script is located. If the cli
    function is being called from a Python shell, this function will raise an
    exception. """
    # pylint: disable=import-outside-toplevel
    import __main__
    if not hasattr(__main__, '__file__'):
        raise MinibossCLIError("Please call miniboss.cli from a Python script")
    return os.path.dirname(os.path.abspath(__main__.__file__))


@cli.command()
@click.option("--exclude", help="Names of services to exclude (comma-separated)")
@click.option("--network-name", help="Network name (generated from group name if not specified)")
@click.option("--timeout", type=int, default=300, help="Timeout for starting a service (seconds)")
def start(exclude, network_name, timeout):
    exclude = exclude.split(",") if exclude else []
    services.start_services(get_main_directory(), exclude, network_name, timeout)


@cli.command()
@click.option("--exclude", help="Names of services to exclude (comma-separated)")
@click.option("--network-name", help="Network name (generated from group name if not specified)")
@click.option("--remove", is_flag=True, default=False, help="Remove container images and network")
@click.option("--timeout", type=int, default=50, help="Timeout for stopping a service (seconds)")
def stop(exclude, network_name, remove, timeout):
    exclude = exclude.split(",") if exclude else []
    services.stop_services(get_main_directory(), exclude, network_name, remove, timeout)

@cli.command()
@click.option("--network-name", help="Network name (generated from group name if not specified)")
@click.option("--timeout", type=int, default=50, help="Timeout for stopping a service (seconds)")
@click.option("--remove", is_flag=True, default=False, help="Remove stopped container")
@click.argument('service')
def reload(service, network_name, timeout, remove):
    services.reload_service(get_main_directory(), service, network_name, remove, timeout)



import logging
import random
import time

import docker
import docker.errors

from miniboss.exceptions import DockerException, ContainerStartException
from miniboss.types import Network

logger = logging.getLogger(__name__)

DIGITS = "0123456789"

_the_docker = None

class DockerClient:

    def __init__(self, lib_client):
        self.lib_client = lib_client

    @classmethod
    def get_client(cls):
        global _the_docker
        if _the_docker is None:
            _the_docker = cls(docker.from_env())
        return _the_docker

    def create_network(self, network_name):
        existing = self.lib_client.networks.list(names=[network_name])
        if existing:
            network = existing[0]
        else:
            network = self.lib_client.networks.create(network_name, driver="bridge")
            logger.info("Created network %s", network_name)
        return network

    def remove_network(self, network_name):
        networks = self.lib_client.networks.list(names=[network_name])
        if networks:
            networks[0].remove()
            logging.info("Removed network %s", network_name)


    def existing_on_network(self, name, network: Network):
        return self.lib_client.containers.list(all=True, filters={'network': network.id,
                                                                  'name': name})

    def build_image(self, build_dir, dockerfile, image_tag):
        try:
            self.lib_client.images.build(tag=image_tag, path=build_dir, dockerfile=dockerfile)
        except docker.errors.BuildError as build_error:
            msg = "Error building image: {}".format(build_error.msg)
            raise DockerException(msg) from None
        except docker.errors.APIError as api_error:
            msg = "Error building image: {}".format(api_error.explanation)
            raise DockerException(msg) from None

    def run_container(self, container_id):
        # The container should be already created but not in state running or starting
        self.lib_client.api.start(container_id)
        # Let's wait a little because the status of the container is
        # not set right away
        time.sleep(1)
        try:
            container = self.lib_client.containers.get(container_id)
        except docker.errors.NotFound:
            raise DockerException(
                "Something went terribly wrong: Could not find container {:s}".format(
                    container_id)) from None
        if container.status != 'running':
            logs = self.lib_client.api.logs(container.id).decode('utf-8')
            raise ContainerStartException(logs, container.name)
        return container

    def check_image(self, tag):
        try:
            self.lib_client.images.get(tag)
        except docker.errors.ImageNotFound:
            pass
        else:
            return
        logger.info("Image %s does not exist, will pull it", tag)
        try:
            self.lib_client.images.pull(tag)
        except docker.errors.APIError as api_error:
            raise DockerException("Could not pull image {} due to API error: {}".format(
                tag, api_error.explanation)) from None


    def run_service_on_network(self,
                               name_prefix,
                               service,  # service: services.Service
                               network: Network):
        container_name = "{:s}-{:s}".format(name_prefix, ''.join(random.sample(DIGITS, 4)))
        networking_config = self.lib_client.api.create_networking_config({
            network.name: self.lib_client.api.create_endpoint_config(aliases=[service.name]),
        })
        host_config=self.lib_client.api.create_host_config(port_bindings=service.ports,
                                                           binds=service.volumes)
        self.check_image(service.image)
        try:
            container = self.lib_client.api.create_container(
                service.image,
                detach=True,
                name=container_name,
                ports=list(service.ports.keys()),
                environment=service.env,
                host_config=host_config,
                networking_config=networking_config,
                volumes=service.volume_def_to_binds(),
                stop_signal=service.stop_signal)
        except docker.errors.ImageNotFound:
            msg = "Image {:s} could not be found; please make sure it exists".format(service.image)
            raise DockerException(msg) from None
        container = self.run_container(container.get('Id'))
        logger.info("Started container id %s for service %s", container.id, service.name)
        return container_name



import json
import pathlib
import logging

from miniboss.exceptions import ContextError

logger = logging.getLogger(__name__)


class _Context(dict):
    filename = ".miniboss-context"

    def save_to(self, directory):
        path = pathlib.Path(directory) / self.filename
        with open(path, 'w') as context_file:
            context_file.write(json.dumps(self))

    def load_from(self, directory):
        path = pathlib.Path(directory) / self.filename
        try:
            with open(path, 'r') as context_file:
                new_data = json.load(context_file)
            self.update(**new_data)
        except FileNotFoundError:
            logger.info("No miniboss context file in %s", directory)

    def remove_file(self, directory):
        path = pathlib.Path(directory) / self.filename
        try:
            path.unlink()
        except FileNotFoundError:
            logger.info("No miniboss context file in %s", directory)

    def extrapolate(self, env_value):
        if not hasattr(env_value, "format"):
            return env_value
        try:
            return env_value.format(**self)
        except KeyError:
            raise ContextError("Could not extrapolate string '{}', existing keys: {}".format(
                env_value, ",".join(self.keys()))) from None
        except ValueError:
            # This happens when there is a type mismatch
            raise ContextError("Could not extrapolate string '{}' due to type mismatch"
                               .format(env_value)) from None
        except IndexError:
            raise ContextError("Only keyword argument extrapolation allowed, violating string: '{}'"
                               .format(env_value)) from None

    def extrapolate_values(self, a_dict):
        return {key: self.extrapolate(value) for key,value in a_dict.items()}

    def _reset(self):
        # Used only for testing
        for key in list(self.keys()):
            self.pop(key)

Context = _Context()



class MinibossException(Exception):
    pass

class ServiceLoadError(MinibossException):
    pass

class ServiceDefinitionError(MinibossException):
    pass

class ServiceAgentException(MinibossException):
    pass

class MinibossCLIError(MinibossException):
    pass

class ContextError(MinibossException):
    pass

class DockerException(MinibossException):
    pass

class ContainerStartException(DockerException):
    def __init__(self, logs, container_name, *args, **kwargs):
        self.logs = logs
        self.container_name = container_name
        super().__init__(*args, **kwargs)

    def __str__(self):
        return "Logs: \n" + self.logs



import threading

from miniboss.service_agent import (ServiceAgent,
                                    Options)

class RunningContext:

    def __init__(self, services_by_name, options: Options):
        super().__init__()
        self.agent_set = {service: ServiceAgent(service, options, self)
                          for name, service in services_by_name.items()}
        self.failed_services = []
        self.processed_services = []
        self.service_pop_lock = threading.Lock()

    @property
    def done(self):
        return self.agent_set == {}

    @property
    def context_failed(self):
        return self.failed_services != []

    @property
    def ready_to_start(self):
        return [x for x in self.agent_set.values() if x.can_start]

    @property
    def ready_to_stop(self):
        return [x for x in self.agent_set.values() if x.can_stop]

    def service_failed(self, failed_service):
        with self.service_pop_lock:
            self.agent_set.pop(failed_service)
            self.failed_services.append(failed_service)
        services_left = list(self.agent_set.keys())
        for service in services_left:
            if failed_service in service.dependencies:
                self.service_failed(service)

    def service_started(self, started_service):
        with self.service_pop_lock:
            self.agent_set.pop(started_service)
            self.processed_services.append(started_service)
            for agent in self.agent_set.values():
                agent.process_service_started(started_service)


    def service_stopped(self, stopped_service):
        with self.service_pop_lock:
            self.agent_set.pop(stopped_service)
            self.processed_services.append(stopped_service)
            for agent in self.agent_set.values():
                agent.process_service_stopped(stopped_service)



import os
import threading
import time
from datetime import datetime
import logging

from miniboss import types
from miniboss.docker_client import DockerClient
from miniboss.context import Context
from miniboss.types import AgentStatus, RunCondition, Actions, Options
from miniboss.exceptions import ServiceAgentException

logger = logging.getLogger(__name__)

def container_env(container):
    env = container.attrs['Config']['Env']
    retval = {}
    for env_line in env:
        key, value = env_line.split('=', 1)
        retval[key] = value
    return retval


def differing_keys(specified, existing):
    """Diff the two environment dictionaries, the first one being the one specified
    in the service def, and the other of an existing container image. Only a one
    way diff; we ignore keys in `existing` that are not in `specified`. We are
    also converting the keys from specified to string because the values from
    existing are always strings anyway. """
    return [key for key,value in specified.items() if str(value) != existing.get(key)]


class ServiceAgent(threading.Thread):

    def __init__(self, service, options: Options, context):
        # service: Service
        # context: RunningContext
        super().__init__()
        self.service = service
        self.options = options
        self.context = context
        self.open_dependencies = service.dependencies[:]
        self.open_dependants = service.dependants[:]
        self.run_condition = RunCondition()
        self.status = AgentStatus.NULL
        self._action = None

    def __repr__(self):
        return "<ServiceAgent service={:s}>".format(self.service.name)

    @property
    def action(self):
        return self._action

    @action.setter
    def action(self, aktion):
        if not aktion in [Actions.START, Actions.STOP]:
            raise ServiceAgentException("Agent action must be one of start or stop")
        self._action = aktion

    @property
    def can_start(self):
        return self.open_dependencies == [] and self.status == AgentStatus.NULL

    @property
    def can_stop(self):
        return self.open_dependants == [] and self.status == AgentStatus.NULL

    @property
    def container_name_prefix(self):
        return "{:s}-{:s}".format(self.service.name, types.group_name)

    def process_service_started(self, service):
        if service in self.open_dependencies:
            self.open_dependencies.remove(service)

    def process_service_stopped(self, service):
        if service in self.open_dependants:
            self.open_dependants.remove(service)

    def build_image(self):
        client = DockerClient.get_client()
        time_tag = datetime.now().strftime("%Y-%m-%d-%H%M")
        image_tag = "{:s}-{:s}".format(self.service.name, time_tag)
        build_dir = os.path.join(self.options.run_dir, self.service.build_from)
        logger.info("Building image with tag %s for service %s from directory %s",
                    image_tag, self.service.name, build_dir)
        client.build_image(build_dir, self.service.dockerfile, image_tag)
        self.run_condition.build_image()
        return image_tag

    def _start_existing(self, existings):
        # pylint: disable=fixme
        # TODO fix this; it should be able to deal with multiple existing
        # containers
        existing = existings[0]
        if existing.status == 'running':
            logger.info("Found running container for %s, not starting a new one",
                        self.service.name)
            self.run_condition.already_running()
            return
        client = DockerClient.get_client()
        if existing.status == 'exited':
            existing_env = container_env(existing)
            diff_keys = differing_keys(self.service.env, existing_env)
            if diff_keys:
                logger.info("Differing env key(s) in existing container for service %s: %s",
                            self.service.name, ",".join(diff_keys))
            start_new = (self.service.always_start_new or
                         self.service.image not in existing.image.tags or
                         bool(diff_keys))
            if not start_new:
                logger.info("There is an existing container for %s, not creating a new one",
                            self.service.name)
                self.run_condition.started()
                client.run_container(existing.id)
                if not self.ping():
                    self._fail()


    def run_image(self): # returns RunCondition
        # pylint: disable=import-outside-toplevel, cyclic-import
        from miniboss.services import Service
        client = DockerClient.get_client()
        self.service.env = Context.extrapolate_values(self.service.env)
        # If there are any running with the name prefix, connected to the same
        # network, skip creating
        existings = client.existing_on_network(self.container_name_prefix,
                                               self.options.network)
        if existings:
            self._start_existing(existings)
            if self.run_condition.state in [RunCondition.STARTED, RunCondition.RUNNING]:
                return
        logger.info("Creating new container for service %s", self.service.name)
        self.service.pre_start()
        if self.service.pre_start.__func__ is not Service.pre_start:
            logger.info("pre_start for service %s ran", self.service.name)
        self.run_condition.pre_started()
        client.run_service_on_network(self.container_name_prefix,
                                      self.service,
                                      self.options.network)

        self.run_condition.started()
        if not self.ping():
            self._fail()
            return
        self.service.post_start()
        self.run_condition.post_started()
        if self.service.post_start.__func__ is not Service.post_start:
            logger.info("post_start for service %s ran", self.service.name)

    def ping(self):
        start = time.monotonic()
        while time.monotonic() - start < self.options.timeout:
            if self.service.ping():
                logger.info("Service %s pinged successfully", self.service.name)
                self.run_condition.pinged()
                return True
            time.sleep(0.1)
        logger.error("Could not ping service with timeout of %d", self.options.timeout)
        return False

    def start_service(self):
        self.action = Actions.START
        self.start()

    def stop_service(self):
        self.action = Actions.STOP
        self.start()

    def run(self):
        if self.action is None:
            self.status = AgentStatus.FAILED
            self.context.service_failed(self.service)
            raise ServiceAgentException("Agent cannot be started without an action set")
        self.status = AgentStatus.IN_PROGRESS
        if self.action == Actions.START:
            self.start_container()
        elif self.action == Actions.STOP:
            self.stop_container()

    def _fail(self):
        self.status = AgentStatus.FAILED
        self.run_condition.fail()
        self.context.service_failed(self.service)
        if RunCondition.START in self.run_condition.actions:
            self._stop_container(remove=True)

    def start_container(self):
        if (self.service.name in self.options.build
            or (self.service.build_from and self.service.image.endswith(':latest'))):
            tag = self.build_image()
            self.service.image = tag
        try:
            self.run_image()
        except Exception: # pylint: disable=broad-except
            logger.exception("Error starting service")
            self._fail()
        if self.run_condition.state == RunCondition.RUNNING:
            logger.info("Service %s started successfully", self.service.name)
            self.status = AgentStatus.STARTED
            self.context.service_started(self.service)

    def _stop_container(self, remove):
        client = DockerClient.get_client()
        existings = client.existing_on_network(self.container_name_prefix,
                                               self.options.network)
        if not existings:
            logging.info("No containers to stop for %s", self.service.name)
        for existing in existings:
            if existing.status == 'running':
                existing.stop(timeout=self.options.timeout)
                logging.info("Stopped container %s", existing.name)
            if remove:
                existing.remove()
                logging.info("Removed container %s", existing.name)

    def stop_container(self):
        self._stop_container(remove=self.options.remove)
        self.status = AgentStatus.STOPPED
        self.context.service_stopped(self.service)



import time
import logging
from collections import Counter, deque
from collections.abc import Mapping

from miniboss import types
from miniboss.docker_client import DockerClient
from miniboss.types import Options, Network
from miniboss.running_context import RunningContext
from miniboss.context import Context
from miniboss.exceptions import MinibossException, ServiceLoadError, ServiceDefinitionError

logging.basicConfig(
    level=logging.INFO,
    format='[%(asctime)s] [%(name)s] %(levelname)s - %(message)s'
)

logger = logging.getLogger(__name__)

KEYCLOAK_PORT = 8090
OSTKREUZ_PORT = 8080
ALLOWED_STOP_SIGNALS = ["SIGINT", "SIGTERM", "SIGKILL", "SIGQUIT"]

class ServiceMeta(type):
    # pylint: disable=too-many-branches
    def __new__(cls, name, bases, attrdict):
        if not bases:
            return super().__new__(cls, name, bases, attrdict)
        if not isinstance(attrdict.get("name"), str) or attrdict["name"] == "":
            raise ServiceDefinitionError(
                "Field 'name' of service class {:s} must be a non-empty string".format(name))
        if not isinstance(attrdict.get("image"), str) or attrdict["image"] == "":
            raise ServiceDefinitionError(
                "Field 'image' of service class {:s} must be a non-empty string".format(name))
        if "ports" in attrdict and not isinstance(attrdict["ports"], Mapping):
            raise ServiceDefinitionError(
                "Field 'ports' of service class {:s} must be a mapping".format(name))
        if "env" in attrdict and not isinstance(attrdict["env"], Mapping):
            raise ServiceDefinitionError(
                "Field 'env' of service class {:s} must be a mapping".format(name))
        if "always_start_new" in attrdict and not isinstance(attrdict["always_start_new"], bool):
            raise ServiceDefinitionError(
                "Field 'always_start_new' of service class {:s} must be a boolean".format(name))
        if "build_from" in attrdict:
            build_dir = attrdict["build_from"]
            if not isinstance(build_dir, str) or build_dir == '':
                raise ServiceDefinitionError(
                    "Field 'build_from' of service class {:s} must be a non-empty string"
                    .format(name))
        if "dockerfile" in attrdict:
            dockerfile = attrdict["dockerfile"]
            if not isinstance(dockerfile, str) or dockerfile == '':
                raise ServiceDefinitionError(
                    "Field 'dockerfile' of service class {:s} must be a non-empty string"
                    .format(name))
        if "stop_signal" in attrdict:
            signal_name = attrdict["stop_signal"]
            if signal_name not in ALLOWED_STOP_SIGNALS:
                raise ServiceDefinitionError(
                    "Stop signal not allowed: {:s}".format(signal_name))
        if "volumes" in attrdict:
            volumes = attrdict["volumes"]
            if isinstance(volumes, list):
                if not all(isinstance(x, str) for x in volumes):
                    raise ServiceDefinitionError(
                        "Volumes have to be defined either as a list of strings or a dict")
            elif isinstance(volumes, dict):
                if not all(isinstance(x, str) for x in volumes.keys()):
                    raise ServiceDefinitionError("Volume definition keys have to be strings")
                for volume in volumes.values():
                    if not isinstance(volume, dict):
                        raise ServiceDefinitionError("Volume definition values have to be dicts")
                    if not isinstance(volume.get('bind'), str):
                        raise ServiceDefinitionError(
                            "Volume definitions have to specify 'bind' key")
        return super().__new__(cls, name, bases, attrdict)


class Service(metaclass=ServiceMeta):
    name = None
    image = ""
    dependencies = []
    ports = {}
    env = {}
    always_start_new = False
    stop_signal = "SIGTERM"
    build_from = None
    dockerfile = "Dockerfile"
    volumes = {}

    # pylint: disable=no-self-use
    def ping(self):
        return True

    def pre_start(self):
        pass

    def post_start(self):
        pass

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.name == other.name

    def __repr__(self):
        return "<miniboss.Service name: {}>".format(self.name)

    def volume_def_to_binds(self):
        if isinstance(self.volumes, dict):
            return [x['bind'] for x in self.volumes.values()]
        return [x.split(':')[1] for x in self.volumes]

def connect_services(services):
    name_counter = Counter()
    for service in services:
        name_counter[service.name] += 1
    multiples = [name for name,count in name_counter.items() if count > 1]
    if multiples:
        raise ServiceLoadError("Repeated service names: {:s}".format(",".join(multiples)))
    all_by_name = {service.name: service for service in services}
    for service in services:
        dependencies = service.dependencies[:]
        for dependency in dependencies:
            if dependency not in all_by_name:
                raise ServiceLoadError(
                    "Dependency {:s} of service {:s} not among services".format(
                        service.name, dependency))
        service.dependencies = [all_by_name[dependency] for dependency in dependencies]
    for service in services:
        service.dependants = [x for x in services if service in x.dependencies]
    return all_by_name

class ServiceCollection:

    def __init__(self):
        self.all_by_name = {}
        self._base_class = Service
        self.running_context = None
        self.excluded = []

    def load_definitions(self):
        services = self._base_class.__subclasses__()
        if len(services) == 0:
            raise ServiceLoadError("No services defined")
        self.all_by_name = connect_services(list(service() for service in services))
        self.check_circular_dependencies()

    def exclude_for_start(self, exclude):
        self.excluded = exclude
        for service in self.all_by_name.values():
            if service.name in exclude:
                continue
            excluded_deps = [dep.name for dep in service.dependencies if dep.name in exclude]
            if excluded_deps:
                raise ServiceLoadError("{:s} is to be excluded, but {:s} depends on it".format(
                    excluded_deps[0], service.name))
        missing = [x for x in exclude if x not in self.all_by_name]
        if missing:
            raise ServiceLoadError("Service{} to be excluded, but not defined: {}".format(
                "s" if len(missing) > 1 else "", ",".join(missing)))
        for name in exclude:
            self.all_by_name.pop(name)


    def exclude_for_stop(self, exclude):
        self.excluded = exclude
        for service_name in exclude:
            service = self.all_by_name[service_name]
            deps_to_be_stopped = [dep.name for dep in service.dependencies
                                  if dep.name not in exclude]
            if deps_to_be_stopped:
                raise ServiceLoadError("{:s} is to be stopped, but {:s} depends on it".format(
                    deps_to_be_stopped[0], service.name))
            self.all_by_name.pop(service_name)


    def check_circular_dependencies(self):
        with_dependencies = [x for x in self.all_by_name.values() if x.dependencies != []]
        # pylint: disable=cell-var-from-loop
        for service in with_dependencies:
            start = service.name
            count = 0
            def go_up_dependencies(checked):
                nonlocal count
                count += 1
                for dependency in checked.dependencies:
                    if dependency.name == start:
                        raise ServiceLoadError("Circular dependency detected")
                    if count == len(self.all_by_name):
                        return
                    go_up_dependencies(dependency)
            go_up_dependencies(service)

    def __len__(self):
        return len(self.all_by_name)

    def check_can_be_built(self, service_name):
        if not service_name in self.all_by_name:
            msg = "No such service: {:s}".format(service_name)
            raise ServiceDefinitionError(msg)
        service = self.all_by_name[service_name]
        if not service.build_from:
            msg = "Service {:s} cannot be built: No build directory specified".format(service.name)
            raise ServiceDefinitionError(msg)

    def start_all(self, options: Options):
        docker = DockerClient.get_client()
        network = docker.create_network(options.network.name)
        options.network.id = network.id
        self.running_context = RunningContext(self.all_by_name, options)
        while not (self.running_context.done or self.running_context.failed_services):
            for agent in self.running_context.ready_to_start:
                agent.start_service()
            time.sleep(0.01)
        failed = []
        if self.running_context.failed_services:
            failed = [x.name for x in self.running_context.failed_services]
            logger.error("Failed to start following services: %s", ",".join(failed))
        return [x for x in self.all_by_name.keys() if x not in failed]


    def stop_all(self, options: Options):
        docker = DockerClient.get_client()
        self.running_context = RunningContext(self.all_by_name, options)
        while not (self.running_context.done or self.running_context.failed_services):
            for agent in self.running_context.ready_to_stop:
                agent.stop_service()
            time.sleep(0.01)
        if options.remove and not self.excluded:
            docker.remove_network(options.network.name)


    def update_for_base_service(self, service_name):
        if service_name not in self.all_by_name:
            raise ServiceLoadError("No such service: {:s}".format(service_name))
        queue = deque()
        queue.append(self.all_by_name[service_name])
        required = []
        while queue:
            service = queue.popleft()
            required.append(service)
            for dependant in service.dependants:
                if dependant not in queue and dependant not in required:
                    queue.append(dependant)
        self.all_by_name = {service.name: service for service in required}

def start_services(maindir, exclude, network_name, timeout):
    if types.group_name is None:
        raise MinibossException(
            "Group name is not set; set it with miniboss.group_name in the main script"
        )
    Context.load_from(maindir)
    collection = ServiceCollection()
    collection.load_definitions()
    collection.exclude_for_start(exclude)
    network_name = network_name or "miniboss-{}".format(types.group_name)
    options = Options(network=Network(name=network_name, id=''),
                      timeout=timeout,
                      remove=False,
                      run_dir=maindir,
                      build=[])
    service_names = collection.start_all(options)
    logger.info("Started services: %s", ", ".join(service_names))
    Context.save_to(maindir)


def stop_services(maindir, exclude, network_name, remove, timeout):
    if types.group_name is None:
        raise MinibossException(
            "Group name is not set; set it with miniboss.group_name in the main script"
        )
    logger.info("Stopping services (excluded: %s)", "none" if not exclude else ",".join(exclude))
    network_name = network_name or "miniboss-{}".format(types.group_name)
    options = Options(network=Network(name=network_name, id=''),
                      timeout=timeout,
                      remove=remove,
                      run_dir=maindir,
                      build=[])
    collection = ServiceCollection()
    collection.load_definitions()
    collection.exclude_for_stop(exclude)
    collection.stop_all(options)
    if remove:
        Context.remove_file(maindir)

# pylint: disable=too-many-arguments
def reload_service(maindir, service, network_name, remove, timeout):
    if types.group_name is None:
        raise MinibossException(
            "Group name is not set; set it with miniboss.group_name in the main script"
        )
    network_name = network_name or "miniboss-{}".format(types.group_name)
    options = Options(network=Network(name=network_name, id=''),
                      timeout=timeout,
                      remove=remove,
                      run_dir=maindir,
                      build=[service])
    stop_collection = ServiceCollection()
    stop_collection.load_definitions()
    stop_collection.check_can_be_built(service)
    stop_collection.update_for_base_service(service)
    stop_collection.stop_all(options)
    # We don't need to do this earlier, as the context is not used by the stop
    # functionality
    Context.load_from(maindir)
    start_collection = ServiceCollection()
    start_collection.load_definitions()
    start_collection.start_all(options)
    Context.save_to(maindir)



import os
from setuptools import setup

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name = "miniboss",
    version = "0.4.0",
    author = "Ulas Turkmen",
    description = "Containerized app testing framework",
    long_description = long_description,
    long_description_content_type='text/markdown',
    install_requires = ["click>7", "docker>4", "furl>2", "requests>2", "attrs>20"],
    tests_require = ["pytest>5.4"],
    packages=['miniboss'],
    url = "https://github.com/afroisalreadyinu/miniboss",
    license = "MIT",
    classifiers = [
        "License :: OSI Approved :: MIT License",
        "Development Status :: 3 - Alpha",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Testing",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers"
    ]
)
